from typing import Optional

class FileMatch:
    """
    Data structure to represent a file's match to the reference string lines. It has three attributes:
    ...

    Attributes
    ----------
    path : str
        a string of the path to the file
    percent_match : float
        the ratio of line similarity between the sequences
    lines : str
        the contents of the file
    """

    path: str
    percent_match: float
    lines: str
    def __new__(cls, path: str, percent_match: float, lines: str) -> FileMatch: ...

def search_for_lines(
    reference_string: str,
    search_path: str,
    max_lines: Optional[int] = None,
    count: Optional[int] = None,
    include_globs: Optional[list[str]] = None,
    exclude_globs: Optional[list[str]] = None,
) -> list[FileMatch]:
    """Search for files with content that most closely match the lines of a reference string."""



from bs4 import BeautifulSoup
from yahoo_fin.stock_info import *

# install requests
# install requests-html

# create symbols list with all securities to be considered
symbols_list = ["KO", "SPY", "VTI", "MUB", "MSEGX", "DBA", "DIS", "V", "JPM", "MSFT", "AAPL"]

df = pd.DataFrame(columns=["Symbol", "Security", "Type", "Category", "Price"])
pd.set_option("display.max_rows", 16)
pd.set_option("display.max_columns", 12)
pd.set_option("display.width", 200)

df["Symbol"] = symbols_list

# loop through all securities
for security_index in range(len(symbols_list)):
    # scrape Yahoo finance security profile page
    url_profile = "https://finance.yahoo.com/quote/%s/profile?p=%s" % (symbols_list[security_index], symbols_list[security_index])
    page_profile = requests.get(url_profile)
    soup_profile = BeautifulSoup(page_profile.content, "html.parser")
    # print(soup_profile.prettify())  # to examine imported profile HTML

    # see if security is a fund, otherwise assume company
    try:  # FUND
        fund_overview = soup_profile.find_all("div", {"class": "Mb(25px)"})[0].find(
            "h3").text.strip()  # look for fund overview section
        if fund_overview == "Fund Overview":
            security_name = soup_profile.find_all("div", {"class": "Mb(20px)"})[0].find("h3").text.strip()
            info_block = soup_profile.find_all("div", {
                "class": "Bdbw(1px) Bdbc($screenerBorderGray) Bdbs(s) H(25px) Pt(10px)"})
            try:  # see if security is a special type of fund
                if info_block[5].find("span", {"class": "Fl(start)"}).text.strip() == "Legal Type":
                    security_type = info_block[5].find("span", {"class": "Fl(end)"}).text.strip()
                else:
                    security_type = "Fund"
            except:
                raise SystemExit("An unknown error occured with the type of a fund security.")
            security_category = info_block[0].find("span", {"class": "Fl(end)"}).text.strip()
            df.at[security_index, "52-Week Change"] = "-"
            df.at[security_index, "Trailing P/E"] = "-"
            df.at[security_index, "Operating Margin"] = "-"
            df.at[security_index, "Quarterly Earnings Growth"] = "-"

            # scrape Yahoo finance security risk page
            url_risk = "https://finance.yahoo.com/quote/%s/risk?p=%s" % (symbols_list[security_index], symbols_list[security_index])
            page_risk = requests.get(url_risk)
            soup_risk = BeautifulSoup(page_risk.content, "html.parser")
            # print(soup_risk.prettify())  # to examine imported profile HTML

            # add the Beta (3Y Monthly) parameter
            if soup_risk.find_all("span", {"data-reactid": "44"})[0].text.strip() == "Beta":
                df.at[security_index, "Beta (3Y Monthly)"] = soup_risk.find_all("span", {"class": "W(39%) Fl(start)", "data-reactid": "46"})[0].text.strip()
            elif soup_risk.find_all("span", {"data-reactid": "74"})[0].text.strip() == "Beta":
                df.at[security_index, "Beta (3Y Monthly)"] = soup_risk.find_all("span", {"class": "W(39%) Fl(start)", "data-reactid": "76"})[0].text.strip()
            else:
                print("The Beta (3Y Monthly) parameter at %s is not in the expected location." % url_risk)

            # add the YTD Return parameter
            if soup_profile.find_all("span", {"data-reactid": "43"})[0].text.strip() == "YTD Return":
                df.at[security_index, "YTD Return"] = soup_profile.find_all("span", {"class": "Fl(end)", "data-reactid": "44"})[0].text.strip()
            elif soup_profile.find_all("span", {"data-reactid": "57"})[0].text.strip() == "YTD Return":
                df.at[security_index, "YTD Return"] = soup_profile.find_all("span", {"class": "Fl(end)", "data-reactid": "58"})[0].text.strip()
            else:
                print("The YTD Return parameter at %s is not in the expected location." % url_profile)

            # add the Annual Report Expense Ratio parameter
            try:
                if soup_profile.find_all("span", {"class": "Mend(5px) Whs(nw)", "data-reactid": "68"})[0].find("span").text.strip() == "Annual Report Expense Ratio (net)":
                    df.at[security_index, "Annual Expense Ratio"] = soup_profile.find_all("span", {"class": "W(20%) D(b) Fl(start) Ta(e)", "data-reactid": "70"})[0].text.strip()
            except:
                try:
                    if soup_profile.find_all("span", {"class": "Mend(5px) Whs(nw)", "data-reactid": "113"})[0].find("span").text.strip() == "Annual Report Expense Ratio (net)":
                        df.at[security_index, "Annual Expense Ratio"] = soup_profile.find_all("span", {"class": "W(20%) D(b) Fl(start) Ta(e)", "data-reactid": "115"})[0].text.strip()
                    else:
                        print("The Annual Report Expense Ratio parameter at %s is not in the expected location." % url_profile)
                except:
                    pass

            print("Success - Fund - %s" % symbols_list[security_index])
            print("Success - Fund Type - %s" % security_type)
            print("Success - Fund Cat - %s" % security_category)
        else:
            print("%s had a Yahoo Finance Profile page that could not be analyzed to find a fund overview section." %
                  symbols_list[security_index])

    except AttributeError:  # COMPANY
        security_name = soup_profile.find_all("div", {"class": "qsp-2col-profile Mt(10px) smartphone_Mt(20px) Lh(1.7)"})[0].find("h3").text.strip()
        security_type = "Company"
        security_category = soup_profile.find_all("span", {"class": "Fw(600)"})[0].text.strip()
        print("Success - Company - %s" % symbols_list[security_index])
        print("Success - Company Type - %s" % security_type)
        print("Success - Company Cat - %s" % security_category)

        # scrape Yahoo finance security statics page
        url_stats = "https://finance.yahoo.com/quote/%s/key-statistics?p=%s" % (symbols_list[security_index], symbols_list[security_index])
        page_stats = requests.get(url_stats)
        soup_stats = BeautifulSoup(page_stats.content, "html.parser")
        # print(soup_stats.prettify())  # to examine imported profile HTML

        # examine 2 locations for 52 Week Change parameter
        try:
            if soup_stats.find_all("span", {"data-reactid": "287"})[0].text.strip() == "52-Week Change":
                df.at[security_index, "52-Week Change"] = soup_stats.find_all("td", {"class": "Fz(s) Fw(500) Ta(end)"})[32].text.strip()
            else:
                print("The 52-Week Change parameter at %s is not in the expected location." % url_stats)
        except IndexError:
            try:
                if soup_stats.find_all("span", {"data-reactid": "292"})[0].text.strip() == "52-Week Change":
                    df.at[security_index, "52-Week Change"] = soup_stats.find_all("td", {"class": "Fz(s) Fw(500) Ta(end)"})[32].text.strip()
                else:
                    print("The 52-Week Change parameter at %s is not in the expected location." % url_stats)
            except IndexError:
                pass

        # add Trailing P/E parameter
        if soup_stats.find_all("span", {"data-reactid": "29"})[1].text.strip() == "Trailing P/E":
            df.at[security_index, "Trailing P/E"] = soup_stats.find_all("td", {"class": "Fz(s) Fw(500) Ta(end)"})[2].text.strip()

        # add Operating Margin parameter
        try:
            if soup_stats.find_all("span", {"data-reactid": "115"})[0].text.strip() == "Operating Margin":
                df.at[security_index, "Operating Margin"] = soup_stats.find_all("td", {"class": "Fz(s) Fw(500) Ta(end)", "data-reactid": "119"})[0].text.strip()
            else:
                print("The Operating Margin parameter at %s is not in the expected location." % url_stats)
        except IndexError:
            try:
                if soup_stats.find_all("span", {"data-reactid": "116"})[0].text.strip() == "Operating Margin":
                    df.at[security_index, "Operating Margin"] = soup_stats.find_all("td", {"class": "Fz(s) Fw(500) Ta(end)", "data-reactid": "120"})[0].text.strip()
                else:
                    print("The Operating Margin parameter at %s is not in the expected location." % url_stats)
            except IndexError:
                pass

        # add Quarterly Earnings Growth parameter
        try:
            if soup_stats.find_all("span", {"data-reactid": "195"})[0].text.strip() == "Quarterly Earnings Growth":
                df.at[security_index, "Quarterly Earnings Growth"] = soup_stats.find_all("td", {"class": "Fz(s) Fw(500) Ta(end)", "data-reactid": "199"})[0].text.strip()
            else:
                print("The Quarterly Earnings Growth parameter at %s is not in the expected location." % url_stats)
        except IndexError:
            try:
                if soup_stats.find_all("span", {"data-reactid": "197"})[0].text.strip() == "Quarterly Earnings Growth":
                    df.at[security_index, "Quarterly Earnings Growth"] = soup_stats.find_all("td", {"class": "Fz(s) Fw(500) Ta(end)", "data-reactid": "201"})[0].text.strip()
                else:
                    print("The Quarterly Earnings Growth parameter at %s is not in the expected location." % url_stats)
            except IndexError:
                pass

        # add Beta (3Y Monthly) parameter
        try:
            if soup_stats.find_all("span", {"data-reactid": "280"})[0].text.strip() == "Beta (3Y Monthly)":
                df.at[security_index, "Beta (3Y Monthly)"] = soup_stats.find_all("td", {"class": "Fz(s) Fw(500) Ta(end)", "data-reactid": "284"})[0].text.strip()
            elif soup_stats.find_all("span", {"data-reactid": "285"})[0].text.strip() == "Beta (3Y Monthly)":
                df.at[security_index, "Beta (3Y Monthly)"] = soup_stats.find_all("td", {"class": "Fz(s) Fw(500) Ta(end)", "data-reactid": "289"})[0].text.strip()
            else:
                print("The Beta (3Y Monthly) parameter at %s is not in the expected location." % url_stats)
        except IndexError:
            print("The Beta (3Y Monthly) parameter at %s is not in the expected location." % url_stats)

        df.at[security_index, "YTD Return"] = "-"
        df.at[security_index, "Annual Expense Ratio"] = "-"

    df.at[security_index, "Security"] = security_name  # assign name to dataframe
    df.at[security_index, "Type"] = security_type  # assign type to dataframe
    df.at[security_index, "Category"] = security_category  # assign category name to dataframe
    df.at[security_index, "Price"] = "%.2f" % (
        get_live_price(symbols_list[security_index]))  # get and assign price to dataframe
    print()

print(df)  # print dataframe


from pynput import mouse

def on_move(x, y):
    print('Pointer moved to {0}'.format(
        (x, y)))

def on_click(x, y, button, pressed):
    print('{0} at {1}'.format(
        'Pressed' if pressed else 'Released',
        (x, y)))
    if not pressed:
        # Stop listener
        return False

def on_scroll(x, y, dx, dy):
    print('Scrolled {0} at {1}'.format(
        'down' if dy < 0 else 'up',
        (x, y)))

# Collect events until released
with mouse.Listener(
        on_move=on_move,
        on_click=on_click,
        on_scroll=on_scroll) as listener:
    listener.join()


from PDFNetPython3 import PDFDoc, Optimizer, SDFDoc
from os import path
from glob import glob

while True:  # repeat for batch processing

    # the base directory where PDF files will be searched
    dir = input('Input path: ')

    # check that the directory is valid
    if not path.isdir(dir):
        raise ValueError('The path input is not a valid path.')

    # search for PDF files recursively or only in base directory
    pdf_files = glob(path.join(dir, '**/*.pdf'), recursive=True)  # recursively
    # pdf_files = glob(path.join(dir, "*.{}".format('pdf')))  # only in the base directory

    # check if there are PDF files
    if not pdf_files:
        raise ValueError(f'No PDF files were found in {dir}')

    # loop through PDF file path names
    for pdf_path in pdf_files:
        # compress PDF file with default settins
        doc = PDFDoc(pdf_path)
        doc.InitSecurityHandler()
        Optimizer.Optimize(doc)
        doc.Save(pdf_path, SDFDoc.e_linearized)
        doc.Close()

        print(f'Compressing {pdf_path}...')

    print('Completed.\n\n')


"""Aldras module containing core objects used across classes"""
import re
import string
import wx
import os
from screeninfo import get_monitors
import math
from operator import add
import uuid
import hashlib
import traceback
import http.client as httplib


def get_system_parameters():
    # get display ranges
    monitors = get_monitors()

    x_indiv = [monitor.x for monitor in monitors]
    widths = [monitor.width for monitor in monitors]
    y_indiv = [monitor.y for monitor in monitors]
    heights = [monitor.height for monitor in monitors]

    x_sum = list(map(add, x_indiv, widths))
    y_sum = list(map(add, y_indiv, heights))

    # get unique hardware id
    uu_id = str(uuid.getnode())
    display_id = ''.join([str(item) for item in
                          x_indiv + y_indiv + widths + heights])  # display configuration id by joining display attributes

    config_id = display_id + uu_id
    config_id = hashlib.sha256(config_id.encode()).hexdigest()  # hash using SHA-256
    config_id = int(math.log(int(config_id, 16), 1.01))  # consolidate by taking log of integer representation

    return (min(x_indiv), max(x_sum)), (min(y_indiv), max(y_sum)), config_id


def check_internet_connection():
    conn = httplib.HTTPConnection("www.google.com", timeout=0.2)
    try:
        conn.request("HEAD", "/")
        conn.close()
        return True
    except Exception as e:
        return False


def coords_of(line):
    """Returns tuple of parsed coordinates from string."""

    try:
        x_coord = re.findall(r'[+-]?\d+', re.findall(r'(?<=\()(.*?)(?=,)', line)[0])[
            0]  # find first integer between '(' and','
    except IndexError:
        x_coord = 0
    try:
        y_coord = re.findall(r'[+-]?\d+', re.findall(r'(?<=,)(.*?)(?=\))', line)[0])[
            0]  # find first integer between ',' and')'
    except IndexError:
        y_coord = 0

    return int(x_coord), int(y_coord)


def eliminate_duplicates(list_with_duplicates):
    """Eliminates duplicates from list"""
    seen = set()
    return [x for x in list_with_duplicates if not (x in seen or seen.add(x))]


def float_in(input_string, return_all=False):
    """Returns parsed float from string."""
    floats = re.findall(r'[-+]?\d*\.\d+|\d+', input_string)
    if not floats:
        output = float(0)
    else:
        output = float(floats[0])
        if len(floats) > 1 and return_all:
            output = [float(indiv_float) for indiv_float in floats]

    return output


def variable_names_in(input_string):
    """Return variable in string between {{~ and ~}} syntax"""
    variables = re.findall(r'(?<={{~)(.*?)(?=~}})', input_string)
    return variables
    # if len(variables) == 1:
    #     return variables[0]
    # else:
    #     if allow_multiple_vars:
    #         return variables
    #     raise ValueError


def assignment_variable_value_in(input_string):
    """Return string after equals sign"""
    return '='.join(input_string.split('=')[1:])


def conditional_operation_in(input_string, operations):
    """Return matching operation between ~}} and ~ syntax"""
    operation_in = re.search(r'(?<=~}})(.*?)(?=~)', input_string).group().lower()
    matching_operations_in = [element for element in operations if element.lower() in operation_in]
    if len(matching_operations_in) == 0:
        return ''
    return matching_operations_in[-1]


def conditional_comparison_in(input_string):
    """Return matching operation between ~ and ~ syntax after variable {{~var~}}"""
    return re.search(r'(?<=~)(.*?)(?=~)', input_string.replace('{{~', '').replace('~}}', '')).group()


def loop_table_data_from(line_text):
    """Return parsed loop table rows and header elements"""

    loop_table_text = line_text[line_text.find('[') + 1:line_text.rfind(']')]  # find text between first '[' and last ']'
    loop_rows = loop_table_text.split("`'''`")  # split based on delimiter
    header_variables = loop_rows[0].split("`'`")  # split based on delimiter

    return loop_rows, header_variables


class PlaceholderTextCtrl(wx.TextCtrl):
    """Placeholder text ctrl."""

    def __init__(self, *args, **kwargs):
        self.default_text = kwargs.pop('placeholder', '')  # remove default text parameter
        wx.TextCtrl.__init__(self, *args, **kwargs)
        self.on_unfocus(None)
        self.Bind(wx.EVT_KEY_DOWN, textctrl_tab_trigger_nav)
        self.Bind(wx.EVT_SET_FOCUS, self.on_focus)
        self.Bind(wx.EVT_KILL_FOCUS, self.on_unfocus)

    def on_focus(self, _):
        self.SetForegroundColour(wx.BLACK)
        if self.GetValue() == self.default_text:
            self.SetValue('')

    def on_unfocus(self, _):
        if self.GetValue().strip() == '':
            self.SetValue(self.default_text)
            self.SetForegroundColour(3 * (120,))


def textctrl_tab_trigger_nav(event):
    """Function to process tab keypresses and trigger panel navigation."""
    if event.GetKeyCode() == wx.WXK_TAB:
        event.EventObject.Navigate()
    event.Skip()


def block_end_index(lines, loop_start_index):
    indent_val = 0  # starting indent value of zero to be increased by the
    loop_end_index = -1  # return index for all lines after loop if no ending bracket is found in the loop below

    # loop through all lines starting from loop until ending bracket is found
    for loop_line_index, loop_line in enumerate(lines[loop_start_index:]):
        line_first_word = loop_line.strip().split(' ')[0].lower()[:6]

        if loop_line.strip() == '}':
            indent_val -= 1
        elif ('if' in line_first_word) and ('{' in loop_line):
            # increase by 1 for new indent block to account for the corresponding ending bracket that should not signal end of block of interest
            indent_val += 1
        elif ('loop' in line_first_word) and ('{' in loop_line):
            # increase by 1 for new indent block to account for the corresponding ending bracket that should not signal end of block of interest
            indent_val += 1

        if indent_val == 0:  # if ending bracket for block of interest is found
            loop_end_index = loop_start_index + loop_line_index
            break  # stop loop

    return loop_end_index


def exception_handler(error_type, value, trace):
    """Handler for all unhandled exceptions."""

    exception = "".join(traceback.format_exception(error_type, value, trace))

    error_dlg = wx.MessageDialog(None, exception, 'An Error Occurred', wx.YES_NO | wx.CANCEL | wx.ICON_ERROR)
    error_dlg.SetYesNoCancelLabels('Report', 'Quit', 'Return')
    error_dlg_response = error_dlg.ShowModal()

    if error_dlg_response == wx.ID_YES:
        # TODO add feedback submission link
        import webbrowser
        webbrowser.open('https://aldras.netlify.com/')
    elif error_dlg_response == wx.ID_NO:
        raise SystemExit()


class CharValidator(wx.Validator):
    """ Validates data as it is entered into the text controls. """

    def __init__(self, flag, parent):
        wx.Validator.__init__(self)
        self.flag = flag
        self.parent = parent
        self.Bind(wx.EVT_CHAR, self.on_char)

    def Clone(self):
        """Required Validator method"""
        return CharValidator(self.flag, self.parent)

    def TransferToWindow(self):
        return True

    def TransferFromWindow(self):
        return True

    def Validate(self, win):  # used when transferred to window
        if self.flag == 'file_name':
            valid = True
            text_ctrl = self.Window
            value = text_ctrl.Value.capitalize()
            proposed_file_name = f'{self.parent.parent.workflow_directory}/{value}.txt'

            if not value.strip():  # if empty string or only spaces
                valid = False
                wx.MessageDialog(None, 'Enter a file name or cancel',
                                 'Invalid file name', wx.OK | wx.ICON_WARNING).ShowModal()

            try:
                if os.path.exists(proposed_file_name):
                    valid = False
                    wx.MessageDialog(None, f'Enter a file name for a file that does not already exist',
                                     'Taken file name', wx.OK | wx.ICON_WARNING).ShowModal()
                else:
                    with open(proposed_file_name, 'w') as _:  # try to create file to validate file name
                        pass
                    os.remove(proposed_file_name)
            except OSError:
                valid = False
                wx.MessageDialog(None, f'Enter a valid file name for your operating system',
                                 'Invalid file name', wx.OK | wx.ICON_WARNING).ShowModal()
            return valid
        else:
            return True

    def on_char(self, event):
        # process character
        keycode = int(event.GetKeyCode())
        if keycode < 256 and not keycode in [8, 127]:  # don't ignore backspace(8) or delete(127)
            key = chr(keycode)
            # return and do not process key if conditions are met
            if self.flag == 'no_alpha' and key in string.ascii_letters:
                return
            if self.flag == 'only_digit' and not (key.isdecimal() or key == '.'):
                return
            if self.flag == 'only_integer' and not key.isdecimal():
                return
            if self.flag == 'coordinate' and not (key.isdecimal() or key == '-'):
                return
            if self.flag == 'file_name':
                invalid_file_characters = ['<', '>', ':', '\"', '\\', '/', '|', '?', '*']
                if key in invalid_file_characters:
                    return
            if self.flag == 'variable_name':
                invalid_variable_characters = ['<', '>', ':', '\"', '\\', '/', '|', '?', '*', '.']
                if key in invalid_variable_characters:
                    return

        event.Skip()
        return


def show_notification(parent, title, message):
    notification = wx.adv.NotificationMessage(title, message, parent)
    notification.SetIcon(wx.Icon('data/aldras.ico', wx.BITMAP_TYPE_ICO))
    notification.Show()


def directory_chooser(parent_window):
    default_directory_dlg = wx.DirDialog(parent_window, 'Choose default save folder', '', wx.DD_DEFAULT_STYLE)
    if default_directory_dlg.ShowModal() == wx.ID_OK:
        dir_path = default_directory_dlg.GetPath()
        default_directory_dlg.SetPath(dir_path)
    else:
        raise SystemExit
    default_directory_dlg.Destroy()

    return dir_path

    # default_directory = os.path.expanduser('~')
    # if os.path.exists(os.path.expanduser('~/Documents')):
    #     default_directory = os.path.expanduser('~/Documents')
    #
    # dlg = wx.FileDialog(parent_window, 'Choose a file', defaultDir=default_directory, wildcard='*.txt')
    # if dlg.ShowModal() == wx.ID_OK:
    #     filename = dlg.GetFilename()
    #     dirname = dlg.GetDirectory()
    #     f = open(os.path.join(dirname, filename), 'r')
    #     parent_window.workflow_name_input.SetValue(f.read())
    #     f.close()
    # dlg.Destroy()



"""Aldras module containing recording objects"""
import wx
from modules.aldras_core import PlaceholderTextCtrl
from modules.aldras_settings import import_settings


def create_record_options(parent_frame, settings_frame=False):

    def record_some_sleep_pressed():
        if not settings_frame:
            some_sleep_thresh.Enable(True)

    def not_some_sleep_pressed():
        if not settings_frame:
            some_sleep_thresh.Enable(False)

    settings = import_settings()

    hbox_options = wx.BoxSizer(wx.HORIZONTAL)

    some_sleep_thresh = 0.2

    # recording pause input
    sleep_sizer = wx.StaticBoxSizer(wx.StaticBox(parent_frame, label='Record pause between actions'), wx.VERTICAL)  # ---------

    record_no_sleep = wx.RadioButton(parent_frame, wx.ID_ANY, 'No pauses')
    record_no_sleep.Bind(wx.EVT_RADIOBUTTON, lambda event: not_some_sleep_pressed())
    sleep_sizer.Add(record_no_sleep, 0, wx.ALL, 4)

    record_all_sleep = wx.RadioButton(parent_frame, wx.ID_ANY, 'All pauses over 0.5')
    record_all_sleep.Bind(wx.EVT_RADIOBUTTON, lambda event: not_some_sleep_pressed())
    sleep_sizer.Add(record_all_sleep, 0, wx.ALL, 4)

    some_sleep_hbox = wx.BoxSizer(wx.HORIZONTAL)
    some_sleep_hbox.AddSpacer(4)
    record_some_sleep = wx.RadioButton(parent_frame, wx.ID_ANY, 'Pauses over')
    record_some_sleep.Bind(wx.EVT_RADIOBUTTON, lambda event: record_some_sleep_pressed())
    some_sleep_hbox.Add(record_some_sleep, 0, wx.ALIGN_CENTER_VERTICAL)

    if settings_frame:
        some_sleep_thresh = wx.TextCtrl(parent_frame, wx.ID_ANY, value=str(settings['Record pause over duration']), size=wx.Size(50, -1),
                                                style=wx.TE_CENTER, name='some_sleep_thresh')
    else:
        some_sleep_thresh = PlaceholderTextCtrl(parent_frame, wx.ID_ANY, placeholder=str(settings['Record pause over duration']), size=wx.Size(50, -1),
                                                style=wx.TE_CENTER, name='some_sleep_thresh')
        some_sleep_thresh.Enable(settings['Record pause'] == 'Pauses over')
    some_sleep_hbox.Add(some_sleep_thresh)

    some_sleep_hbox.Add(wx.StaticText(parent_frame, label='  sec   '), 0, wx.ALIGN_CENTER_VERTICAL)
    sleep_sizer.Add(some_sleep_hbox, 0, wx.BOTTOM, 5)

    hbox_options.Add(sleep_sizer, 0, wx.ALL, 5)
    # ------------------------------------------------------------------------------------------------------
    hbox_options.AddSpacer(5)

    # recording method input
    record_mthd = wx.RadioBox(parent_frame, label='Method', choices=['Overwrite', 'Append'], majorDimension=1, style=wx.RA_SPECIFY_COLS, name='Record method')
    record_mthd.SetItemToolTip(0, 'Erase existing data')
    record_mthd.SetItemToolTip(1, 'Add to end of existing data')
    hbox_options.Add(record_mthd, 0, wx.ALL, 5)

    return hbox_options


import os
from os.path import basename
import shutil
from time import sleep
from zipfile import ZipFile

import gunicorn
from flask import Flask, abort, after_this_request, render_template, request, send_from_directory
from flask_cors import CORS
from flask_limiter import Limiter
from flask_limiter.util import get_remote_address
from werkzeug.utils import secure_filename

import doc_find_replace

app = Flask(__name__)

limiter = Limiter(app, key_func=get_remote_address)

CORS(app)


def eprint(*args, **kwargs):
    if __name__ == "__main__":
        """
        Printing to stderr so visible when using flask dev server
        """
        import sys
        from pprint import pprint as pp

        pp(*args, sys.stderr, **kwargs)
    else:
        print(*args, **kwargs)


def clear_folder(dir: str):
    if not os.path.exists(dir):
        os.makedirs(dir)

    for file in os.listdir(dir):
        try:
            os.remove(os.path.join(dir, file))
        except PermissionError as error:
            print(error)


@app.route("/", methods=["GET", "POST"])
@limiter.limit("20/day")
def upload_file():
    if request.method == "POST":
        upload_dir = "uploads"
        output_dir = "created"

        # Delete all working files
        clear_folder(upload_dir)
        clear_folder(output_dir)

        template_file = request.files.get("template_file")
        replacements_file = request.files.get("replacements_file")

        eprint(f"template_file : {template_file}")
        eprint(f"replacements_file : {replacements_file}")

        if template_file and replacements_file:
            template_fn = secure_filename(template_file.filename)
            template_file.save(os.path.join(upload_dir, template_fn))

            output_base_fn = secure_filename(request.form.get("output_base_fn", "")) or template_fn.replace(".docx", "")

            replacements_fn = secure_filename(replacements_file.filename)
            replacements_file.save(os.path.join(upload_dir, replacements_fn))

            eprint((template_fn, replacements_fn, output_base_fn))

            output_file_paths = doc_find_replace.batch_replace(
                replacements_csv=os.path.join(upload_dir, replacements_fn),
                template_docx=os.path.join(upload_dir, template_fn),
                output_dir=output_dir,
                output_base_fn=output_base_fn,
                output_filetype=".docx",
            )

            print(f"output_file_paths : {output_file_paths}")
            print(os.listdir(output_dir))

            # output_zip_path = os.path.join(output_dir, "generated_documents.zip")
            # with ZipFile(output_zip_path, "w") as zip_obj:
            #     for generated_doc_path in output_file_paths:
            #         zip_obj.write(generated_doc_path,
            #         basename(generated_doc_path))

            output_zip_fn = request.form.get("output_zip_fn", "generated_documents.zip")
            shutil.make_archive(root_dir="created", format="zip", base_name=output_zip_fn.replace(".zip", ""))

            # Delete all working files after request
            @after_this_request
            def clear__files(response):
                clear_folder(upload_dir)
                clear_folder(output_dir)
                return response

            # Download the file
            try:
                return send_from_directory("", output_zip_fn, as_attachment=True)
            except FileNotFoundError:
                abort(404)

    return render_template("main.html")


if __name__ == "__main__":
    # Quick test configuration. Please use proper Flask configuration options
    # in production settings, and use a separate file or environment variables
    # to manage the secret key!
    app.secret_key = os.urandom(12).hex()
    app.config["SESSION_TYPE"] = "filesystem"

    app.debug = True
    app.run()



print("Hello World A")
print("Hello World B")

print("Hello World 4")
print("Hello World C")

print("Hello World D")
print("Hello World E")
print("Hello World F")
print("Hello World G")
print("Hello World H")
print("Hello World H")
print("Hello World H")
print("Hello World H")
print("Hello World H")
print("Hello World H")
print("Hello World H")
print("Hello World H")
print("Hello World H")
print("Hello World H")
print("Hello World H")
print("Hello World H")
print("Hello World H")
print("Hello World H")
print("Hello World H")
print("Hello World H")



import threading
import pyperclip
import re
import wx
import wx.adv
import wx.grid
import wx.lib.expando
import wx.lib.scrolledpanel
from pynput import keyboard


def variable_names_in(input_string):
    """Return variable in string between {{~ and ~}} syntax"""
    variables = re.findall(r'(?<={{~)(.*?)(?=~}})', input_string)
    return variables


def change_font(widget, size=None, family=None, style=None, weight=None, color=None):
    # set default parameters
    size = size if size is not None else 9
    family = family if family is not None else wx.DEFAULT
    style = style if style is not None else wx.NORMAL
    weight = weight if weight is not None else wx.NORMAL

    widget.SetFont(wx.Font(size, family, style, weight))

    if color is not None:
        widget.SetForegroundColour(color)


class App(wx.App):
    """Main App to Contain Data."""

    def __init__(self, macros):
        wx.App.__init__(self)

        # check that macros array elements are strings
        for macro in macros:
            for element in macro:
                if not isinstance(element, str):
                    print(element)
                    wx.MessageDialog(None,
                                     f'Macro Data Error\n\nThe macro data one or more macro triggers and/or content that are not strings.\nPlease double check the provided macro data, especially the element \'{element}\'.',
                                     'Macro Data Must in String Format', wx.ICON_ERROR).ShowModal()
                    wx.Exit()

        MainFrame(None, macros)


class MainFrame(wx.Frame):
    """Main frame to close application."""

    def __init__(self, parent, macros):
        wx.Frame.__init__(self, parent, title=f'Salesforce Macro Automation', name='main_frame')
        self.SetIcon(wx.Icon('data/salesforce.ico', wx.BITMAP_TYPE_ICO))  # assign icon
        self.SetBackgroundColour('white')  # set background color
        self.macros = macros
        self.monitor_thread = MonitorThread(self, macros)
        self.macro_being_executed = False
        self.keyboard_controller = keyboard.Controller()

        self.workflow_panel = wx.Panel(self)

        # set parameters
        self.margin_y = 25
        self.margin_x = 60
        self.padding_y = 25

        self.vbox = wx.BoxSizer(wx.VERTICAL)  # ------------------------------------------------------------------------

        # add rescaled logo image
        png = wx.Image('data/sf2.png', wx.BITMAP_TYPE_PNG).Scale(150, 150,
                                                                         quality=wx.IMAGE_QUALITY_HIGH)
        self.logo_img = wx.StaticBitmap(self.workflow_panel, wx.ID_ANY, wx.Bitmap(png))
        self.vbox.Add(self.logo_img, 0, wx.ALIGN_CENTER_HORIZONTAL | wx.SOUTH, self.padding_y)

        # add program name text
        self.program_name = wx.StaticText(self.workflow_panel, label='Salesforce Dynamic Macro Automation')
        change_font(self.program_name, size=17, color=3 * (60,))
        self.vbox.Add(self.program_name, 0, wx.ALIGN_CENTER_HORIZONTAL | wx.SOUTH, self.padding_y*3/4)

        # add main instructions text
        self.instructions = wx.StaticText(self.workflow_panel, label='Currently listening for macro triggers...')
        change_font(self.instructions, size=13, color=3 * (60,))
        self.vbox.Add(self.instructions, 0, wx.ALIGN_CENTER_HORIZONTAL | wx.SOUTH, self.padding_y*3)

        # add credit text
        self.credit = wx.StaticText(self.workflow_panel, label='Noah Baculi 2021')
        change_font(self.credit, size=7, color=3 * (60,))
        self.vbox.Add(self.credit, 0, wx.ALIGN_CENTER_HORIZONTAL | wx.SOUTH, self.padding_y)
        # --------------------------------------------------------------------------------------------------------------

        # add outer sizer
        self.vbox_outer = wx.BoxSizer(wx.VERTICAL)
        self.vbox_outer.AddStretchSpacer()
        self.vbox_outer.AddSpacer(self.margin_y)  # north margin
        self.vbox_outer.Add(self.vbox, 0, wx.ALIGN_CENTER_HORIZONTAL | wx.EAST | wx.WEST, self.margin_x)

        # display frame
        self.workflow_panel.SetSizerAndFit(self.vbox_outer)
        self.vbox_outer.SetSizeHints(self)
        self.Center()
        self.Show()

        self.monitor_thread.start()

    def execute_macro(self, trigger):
        if not self.macro_being_executed:
            self.macro_being_executed = True
            output = self.macros[trigger]
            variable_names = variable_names_in(output)

            if variable_names:  # dynamic macro, do not output text
                var_dlg = VariablesDialog(self, variable_names)
                var_dlg.ShowModal()

                for variable in var_dlg.variables.items():
                    output = output.replace(f'{{{{~{variable[0]}~}}}}', variable[1])

            else:  # static macro, delete trigger and output text
                trigger_output_keys = trigger
                for key_no_output in ['pause', 'enter', 'tab', 'home', 'page_up', 'page_down']:
                    trigger_output_keys = trigger_output_keys.replace('pause', '')
                for ii in range(len(trigger_output_keys)):
                    self.keyboard_controller.press(keyboard.Key.backspace)
                    self.keyboard_controller.release(keyboard.Key.backspace)
                    
                self.keyboard_controller.type(output)

            # print(output)
            pyperclip.copy(output)

        self.macro_being_executed = False


class MonitorThread(threading.Thread):
            def __init__(self, thread_parent, macros):
                """Init Worker Thread Class."""
                threading.Thread.__init__(self, daemon=True)
                self.parent = thread_parent
                self.macros = macros
                self.released_keys = ''
                self.keyboard_controller = keyboard.Controller()

            def run(self):
                """Run worker thread."""

                def process_keystroke(key):
                    """
                    Process keystroke press or release for keyboard listener for ListenerThread instances.

                    key: parameter passed by pynput listener identifying the key
                    """

                    key = key.process_numpad()

                    # strip single quotes and lower
                    key = str(key).strip('\'').lower()

                    # eliminate 'key.' and make substitutions for backslash (\), single quote ('), and right ctrl (should be replaced with left)
                    key = key.replace('key.', '').replace('\\\\', '\\').replace('\"\'\"', '\'')

                    self.released_keys += (key)
                    self.released_keys = self.released_keys[-20:]  # limit stored released keys to 20
                    print(self.released_keys)

                    for trigger in self.macros.keys():
                        if trigger in self.released_keys:
                            self.released_keys = ''  # clear stored keys
                            wx.CallAfter(self.parent.execute_macro, trigger)

                def process_numpad(key):
                    # replace numberpad virtual keycodes with strings of the number digits
                    for number in range(10):  # loop through 0-9 and replace keycodes starting at <96>
                        key = str(key).replace(f'<{96 + number}>', str(number))

                    # replace numberpad period with string
                    key = str(key).replace('<110>', '.')
                    return key

                self.key_listener = keyboard.Listener(on_press=process_keystroke)
                try:
                    self.key_listener.join()
                except RuntimeError:
                    self.key_listener.start()

            def abort(self):
                """Abort worker thread."""
                # Method for use by main thread to signal an abort
                self.key_listener.stop()


class VariablesDialog(wx.Dialog):
    """Dialog to verify license."""

    def __init__(self, parent, variable_names):
        wx.Dialog.__init__(self, parent, title=f'Variables', name='variables_dialog', style=wx.STAY_ON_TOP)
        self.SetIcon(wx.Icon('data/salesforce.ico', wx.BITMAP_TYPE_ICO))
        self.SetBackgroundColour('white')
        self.parent = parent
        self.variable_names = variable_names
        self.variables = {}
        self.Bind(wx.EVT_CLOSE, self.OnClose)
        self.license_panel = wx.Panel(self)

        self.margin_y = 10
        self.margin_x = 15

        self.vbox = wx.BoxSizer(wx.VERTICAL)  # ------------------------------------------------------------------------

        self.hbox_logo_name_version = wx.BoxSizer(wx.HORIZONTAL)

        # add rescaled logo image
        png = wx.Image('data/salesforce.png', wx.BITMAP_TYPE_PNG).Scale(60, 60, quality=wx.IMAGE_QUALITY_HIGH)
        self.logo_img = wx.StaticBitmap(self, wx.ID_ANY, wx.Bitmap(png))
        self.hbox_logo_name_version.Add(self.logo_img, 0, wx.ALIGN_CENTER_VERTICAL | wx.EAST, 10)
        self.vbox_name_version = wx.BoxSizer(wx.VERTICAL)

        # add program name text
        self.program_name = wx.StaticText(self, label='Variables')
        change_font(self.program_name, size=16, color=3 * (20,))
        self.vbox_name_version.Add(self.program_name, 0, wx.ALIGN_CENTER_HORIZONTAL)

        # add program version text
        self.program_version = wx.StaticText(self, label=f'2021')
        change_font(self.program_version, size=10, style=wx.ITALIC, color=3 * (80,))
        self.vbox_name_version.Add(self.program_version, 0, wx.ALIGN_CENTER_HORIZONTAL)

        self.hbox_logo_name_version.Add(self.vbox_name_version, 0, wx.ALIGN_CENTER_VERTICAL)
        self.vbox.Add(self.hbox_logo_name_version, 0, wx.ALIGN_CENTER_HORIZONTAL | wx.SOUTH, 10)

        for var_name in self.variable_names:
            # add input field for the license key
            self.hbox_key_input = wx.BoxSizer(wx.HORIZONTAL)

            self.hbox_key_input.Add(wx.StaticText(self.license_panel, label=var_name), 0,
                                    wx.ALIGN_CENTER_VERTICAL | wx.WEST, 20)

            self.hbox_key_input.AddSpacer(10)

            self.license_key_input = wx.TextCtrl(self.license_panel, wx.ID_ANY, size=(280, -1),
                                                         style=wx.TE_PROCESS_ENTER | wx.TE_CENTRE, name=var_name)
            self.Bind(wx.EVT_TEXT_ENTER, self.activate, self.license_key_input)
            self.hbox_key_input.Add(self.license_key_input, 0, wx.ALIGN_CENTER_VERTICAL | wx.EAST, 20)
            self.vbox.Add(self.hbox_key_input, 0, wx.ALIGN_CENTER_HORIZONTAL | wx.SOUTH, 25)

        # add buttons
        self.button_array = wx.StdDialogButtonSizer()

        self.button_array.AddStretchSpacer()
        self.close_btn = wx.Button(self.license_panel, label='Cancel')
        self.close_btn.Bind(wx.EVT_BUTTON, self.OnClose)
        self.button_array.Add(self.close_btn)
        self.button_array.AddSpacer(5)
        self.activate_btn = wx.Button(self.license_panel, label='Copy to Clipboard')
        self.activate_btn.Bind(wx.EVT_BUTTON, self.activate)
        self.button_array.Add(self.activate_btn)

        self.vbox.Add(self.button_array, 0, wx.EXPAND | wx.SOUTH, 10)

        self.vbox_outer = wx.BoxSizer(wx.VERTICAL)
        self.vbox_outer.AddSpacer(self.margin_y)  # north margin
        self.vbox_outer.Add(self.vbox, 0, wx.ALIGN_CENTER_HORIZONTAL | wx.EAST | wx.WEST, self.margin_x)

        # display frame
        self.license_panel.SetSizerAndFit(self.vbox_outer)
        self.vbox_outer.SetSizeHints(self)
        # self.activate_btn.SetFocus()
        self.Center()
        self.Show()

    def OnClose(self, _):
        self.Destroy()

    def activate(self, _):
        print('Activated')
        for var_name in self.variable_names:
            var_value = self.FindWindowByName(var_name).GetValue()
            self.variables[var_name] = var_value
        self.Destroy()


class PlaceholderTextCtrl(wx.TextCtrl):
    """Placeholder text ctrl."""

    def __init__(self, *args, **kwargs):
        self.default_text = kwargs.pop('placeholder', '')  # remove default text parameter
        wx.TextCtrl.__init__(self, *args, **kwargs)
        self.on_unfocus(None)

        def textctrl_tab_trigger_nav(event):
            """Function to process tab keypresses and trigger panel navigation."""
            if event.GetKeyCode() == wx.WXK_TAB:
                event.EventObject.Navigate()
            event.Skip()

        self.Bind(wx.EVT_KEY_DOWN, textctrl_tab_trigger_nav)
        self.Bind(wx.EVT_SET_FOCUS, self.on_focus)
        self.Bind(wx.EVT_KILL_FOCUS, self.on_unfocus)

    def on_focus(self, _):
        self.SetForegroundColour(wx.BLACK)
        if self.GetValue() == self.default_text:
            self.SetValue('')

    def on_unfocus(self, _):
        if self.GetValue().strip() == '':
            self.SetValue(self.default_text)
            self.SetForegroundColour(3 * (120,))




if __name__ == '__main__':
    ### MACROS
    # Add macros here in a list of lists with {{~VAR~}} formatting for any dynamic variables.
    # Limit macro triggers to less than 20 alphanumeric characters and avoid modifier keys or words
    macros = {'intropause':"Hi there!"}

    app = App(macros)
    app.MainLoop()


"""Aldras module containing execution objects"""
import wx
from modules.aldras_core import PlaceholderTextCtrl
from modules.aldras_settings import import_settings


def create_execute_options(parent_frame, settings_frame=False):
    def checkbox_pause_pressed():
        if not settings_frame:
            parent_frame.execute_pause_input.Enable(parent_frame.checkbox_pause.GetValue())

    def checkbox_mouse_dur_pressed():
        if not settings_frame:
            parent_frame.execute_mouse_dur_input.Enable(parent_frame.checkbox_mouse_dur.GetValue())

    def checkbox_type_interval_pressed():
        if not settings_frame:
            parent_frame.execute_type_interval_input.Enable(parent_frame.checkbox_type_interval.GetValue())

    def textctrl(placeholder_val, name):
        # TODO implement numeric validator
        if settings_frame:
            return wx.TextCtrl(parent_frame, wx.ID_ANY, value=str(placeholder_val), size=wx.Size(50, -1), style=wx.TE_CENTER, name=name)
        else:
            return PlaceholderTextCtrl(parent_frame, wx.ID_ANY, placeholder=str(placeholder_val), size=wx.Size(50, -1), style=wx.TE_CENTER, name=name)

    settings = import_settings()

    vbox = wx.BoxSizer(wx.VERTICAL)
    vbox.AddSpacer(10)

    # execution pause input
    hbox_pause = wx.BoxSizer(wx.HORIZONTAL)  # --------------------------------------------------------

    parent_frame.checkbox_pause = wx.CheckBox(parent_frame, label=' Pause between commands: ', name='Execute pause between commands checked')
    parent_frame.checkbox_pause.SetValue(settings['Execute pause between commands checked'])
    parent_frame.checkbox_pause.Bind(wx.EVT_CHECKBOX, lambda event: checkbox_pause_pressed())
    hbox_pause.Add(parent_frame.checkbox_pause, 0, wx.ALIGN_CENTER_VERTICAL)

    parent_frame.execute_pause_input = textctrl(settings['Execute pause between commands'], 'Execute pause between commands')
    hbox_pause.Add(parent_frame.execute_pause_input, 0, wx.ALIGN_CENTER_VERTICAL)

    hbox_pause.Add(wx.StaticText(parent_frame, label='  sec'), 0, wx.ALIGN_CENTER_VERTICAL)
    vbox.Add(hbox_pause, 0, wx.EAST | wx.WEST, 10)
    # ------------------------------------------------------------------------------------------------------

    vbox.AddSpacer(20)

    # Mouse duration input
    hbox_mouse_dur = wx.BoxSizer(wx.HORIZONTAL)  # ----------------------------------------------------

    parent_frame.checkbox_mouse_dur = wx.CheckBox(parent_frame, label=' Mouse command duration: ', name='Execute mouse command duration checked')
    parent_frame.checkbox_mouse_dur.SetValue(settings['Execute mouse command duration checked'])
    parent_frame.checkbox_mouse_dur.Bind(wx.EVT_CHECKBOX, lambda event: checkbox_mouse_dur_pressed())
    hbox_mouse_dur.Add(parent_frame.checkbox_mouse_dur, 0, wx.ALIGN_CENTER_VERTICAL)

    parent_frame.execute_mouse_dur_input = textctrl(settings['Execute mouse command duration'], 'Execute mouse command duration')
    hbox_mouse_dur.Add(parent_frame.execute_mouse_dur_input, 0, wx.ALIGN_CENTER_VERTICAL)

    hbox_mouse_dur.Add(wx.StaticText(parent_frame, label='  sec'), 0, wx.ALIGN_CENTER_VERTICAL)
    vbox.Add(hbox_mouse_dur, 0, wx.EAST | wx.WEST, 10)
    # ------------------------------------------------------------------------------------------------------

    vbox.AddSpacer(20)

    # Text type interval duration input
    hbox_type_interval = wx.BoxSizer(wx.HORIZONTAL)  # ------------------------------------------------

    parent_frame.checkbox_type_interval = wx.CheckBox(parent_frame, label=' Interval between text character outputs: ', name='Interval between text character outputs checked')
    parent_frame.checkbox_type_interval.SetValue(settings['Interval between text character outputs checked'])
    parent_frame.checkbox_type_interval.Bind(wx.EVT_CHECKBOX, lambda event: checkbox_type_interval_pressed())
    hbox_type_interval.Add(parent_frame.checkbox_type_interval, 0, wx.ALIGN_CENTER_VERTICAL)

    parent_frame.execute_type_interval_input = textctrl(settings['Interval between text character outputs'], 'Interval between text character outputs')
    hbox_type_interval.Add(parent_frame.execute_type_interval_input, 0, wx.ALIGN_CENTER_VERTICAL)

    hbox_type_interval.Add(wx.StaticText(parent_frame, label='  sec'), 0, wx.ALIGN_CENTER_VERTICAL)
    vbox.Add(hbox_type_interval, 0, wx.EAST | wx.WEST, 10)
    # ------------------------------------------------------------------------------------------------------

    checkbox_pause_pressed()
    checkbox_mouse_dur_pressed()
    checkbox_type_interval_pressed()

    return vbox


import inspect
import time
import unittest
from time import perf_counter

import busca_py as busca


def example_usage():
    reference_file_path = "./sample_dir_hello_world/file_1.py"
    with open(reference_file_path, "r") as file:
        reference_string = file.read()

    # Perform search with required parameters
    all_file_matches = busca.search_for_lines(
        reference_string=reference_string,
        search_path="./sample_dir_hello_world",
    )

    # File matches are returned in descending order of percent match
    closest_file_match = all_file_matches[0]
    assert closest_file_match.path == reference_file_path
    assert closest_file_match.percent_match == 1.0
    assert closest_file_match.lines == reference_string

    # Perform search for top 5 matches with additional filters
    # to speed up runtime by skipping files that will not match
    relevant_file_matches = busca.search_for_lines(
        reference_string=reference_string,
        search_path="./sample_dir_hello_world",
        max_lines=10_000,
        include_globs=["*.py"],
        count=5,
    )

    assert len(relevant_file_matches) < len(all_file_matches)

    # Create new file match object
    new_file_match = busca.FileMatch("file/path", 1.0, "file\ncontent")


class TestSignatures(unittest.TestCase):
    def test_module_contains_functions(self):
        expected_functions = {"search_for_lines"}
        module_attributes = set(dir(busca))
        self.assertTrue(expected_functions.issubset(module_attributes))

    def test_non_empty_search_function_signature(self):
        self.assertTrue(inspect.signature(busca.search_for_lines).parameters.items())

    def test_non_empty_file_match_class_signature(self):
        self.assertTrue(inspect.signature(busca.FileMatch).parameters.items())


class TestSearchResults(unittest.TestCase):
    def setUp(self):
        with open("./sample_dir_hello_world/file_1.py", "r") as file:
            ref_str = file.read()
        self.search = busca.search_for_lines(
            reference_string=ref_str,
            search_path="./",
            max_lines=10000,
            count=5,
            include_globs=["*.py"],
        )

    def test_first_result(self):
        file_match = self.search[0]

        expected_lines = 'print("Hello World 1")\nprint("Hello World 2")\n\n\nprint("Hello World 3")\nprint("Hello World 4")\n\nprint("Hello World 5")\nprint("Hello World 6")'

        self.assertEqual(file_match.path, "./sample_dir_hello_world/file_1.py")
        self.assertEqual(file_match.percent_match, 1.0)
        self.assertEqual(file_match.lines, expected_lines)

    def test_third_result(self):
        file_match = self.search[2]

        expected_lines = '\n\nprint("Hello World 1")\n\nprint("Hello World 3")\n'

        self.assertEqual(file_match.path, "./sample_dir_hello_world/nested_dir/sample_python_file_3.py")
        self.assertEqual(file_match.percent_match, 0.4285714328289032)
        self.assertEqual(file_match.lines, expected_lines)


class TestSearchDuration(unittest.TestCase):
    def setUp(self):
        with open("./sample_dir_hello_world/file_1.py", "r") as file:
            self.ref_str = file.read()

    def test_no_globs(self):
        t1 = perf_counter()
        _ = busca.search_for_lines(
            reference_string=self.ref_str,
            search_path="./",
            max_lines=10000,
            count=5,
        )
        duration = perf_counter() - t1
        self.assertLess(duration, 5)

    def test_only_py_files(self):
        t1 = perf_counter()
        _ = busca.search_for_lines(
            reference_string=self.ref_str,
            search_path="./",
            max_lines=10000,
            count=5,
            include_globs=["*.py"],
        )
        duration = perf_counter() - t1
        self.assertLess(duration, 5)


if __name__ == "__main__":
    # unittest.main()
    # example_usage()

    # search_path = "./sample_dir_mix"
    # search_path = "./"
    search_path = "/Users/noahbaculi/github/profiles/wt_2/profiles"

    with open(
        "/Users/noahbaculi/github/profiles/wt_2/profiles/carium/event_handlers/add_to_group_eventhandler.py", "r"
    ) as file:
        reference_string_3 = file.read()
    t3 = time.perf_counter()
    _ = busca.search_for_lines(reference_string_3, search_path, max_lines=10_000)
    print(f"3 = {time.perf_counter() - t3:.2f} seconds")

    with open(
        "/Users/noahbaculi/github/profiles/wt_2/profiles/carium/event_handlers/elation_object_handler.py", "r"
    ) as file:
        reference_string = file.read()
    t1 = time.perf_counter()
    all_file_matches = busca.search_for_lines(reference_string, search_path, max_lines=10_000)
    print(f"1 = {time.perf_counter() - t1:.2f} seconds")

    with open("/Users/noahbaculi/github/profiles/wt_2/profiles/carium/event_handlers/pathway_handler.py", "r") as file:
        reference_string_2 = file.read()
    t2 = time.perf_counter()
    _ = busca.search_for_lines(reference_string_2, search_path, max_lines=10_000)
    print(f"2 = {time.perf_counter() - t2:.2f} seconds")

    with open(
        "/Users/noahbaculi/github/profiles/wt_2/profiles/carium/event_handlers/add_to_group_eventhandler.py", "r"
    ) as file:
        reference_string_3 = file.read()
    t3 = time.perf_counter()
    _ = busca.search_for_lines(reference_string_3, search_path, max_lines=10_000)
    print(f"3 = {time.perf_counter() - t3:.2f} seconds")



"""Aldras module containing settings objects"""
import json
import os

import wx

from modules.aldras_core import CharValidator, directory_chooser

settings_possibilities = {
    'Number of recent workflows displayed': [str(ii) for ii in list(range(0, 6))],  # zero to five
    'Freeze method': ['None', 'Click', 'Ctrl', 'Click or ctrl'],
    'Number of hotkeys': ['2', '3', '4'],
    'Notifications': ['Banners', 'Windows'],
    'Record pause': ['No pauses', 'All pauses over 0.5', 'Pauses over'],
    'Record method': ['Overwrite', 'Append'],
}


def validate_settings(settings_unvalidated):
    # factory settings for reference comparison (double quotes used rather than single quote convention due to desire to allow copy-paste between this dictionary and the .json file)
    factory_settings = {
        "Number of recent workflows displayed": 3,
        "Workflow folder": "",
        "Freeze method": "Ctrl",
        "Number of hotkeys": 3,
        "Large lines number": 100,
        "Notifications": "Banners",
        "Record pause": "Pauses over",
        "Record pause over duration": 5.0,
        "Record method": "Append",
        "Execute pause between commands": 1.0,
        "Execute pause between commands checked": "True",
        "Execute mouse command duration": 0.5,
        "Execute mouse command duration checked": "True",
        "Interval between text character outputs": 0.02,
        "Interval between text character outputs checked": "True"
    }

    # settings validation lambda functions
    settings_validation = {
        'Number of recent workflows displayed': lambda x: str(x) in settings_possibilities[
            'Number of recent workflows displayed'],
        "Workflow folder": lambda dir_path: os.path.exists(dir_path),
        'Freeze method': lambda x: x.lower() in [y.lower() for y in settings_possibilities['Freeze method']],
        'Number of hotkeys': lambda x: 2 <= x <= 4,
        'Large lines number': lambda x: 15 <= x <= 200,
        'Notifications': lambda x: x.lower() in [y.lower() for y in settings_possibilities['Notifications']],
        'Record pause': lambda x: x.lower() in [y.lower() for y in settings_possibilities['Record pause']],
        'Record pause over duration': lambda x: x > 0,
        'Record method': lambda x: x.lower() in [y.lower() for y in settings_possibilities['Record method']],
        'Execute pause between commands': lambda x: x > 0,
        'Execute pause between commands checked': lambda x: x in [True, False],
        'Execute mouse command duration': lambda x: x > 0,
        'Execute mouse command duration checked': lambda x: x in [True, False],
        'Interval between text character outputs': lambda x: x > 0,
        'Interval between text character outputs checked': lambda x: x in [True, False]
    }

    settings = dict()
    for key in factory_settings:  # loop through the factory settings and attempt to parse imported setting if available

        # determine type of factory settings to cast imported settings
        if factory_settings[key] in ['True', 'False']:
            cast_type = bool
        else:
            cast_type = type(factory_settings[key])

        try:
            if settings_validation[key](
                    cast_type(settings_unvalidated[key])):  # if the cast imported setting is validated
                settings[key] = cast_type(settings_unvalidated[key])  # set equal to the cast imported setting

                if key != 'Workflow folder' and isinstance(settings[key],
                                                           str):  # do not modify captilization of workflow folder path (just for aesthetic reasons)
                    settings[key] = settings[key].capitalize()  # capitalize setting if string
            else:
                raise ValueError
        except (KeyError, ValueError):
            settings[key] = cast_type(factory_settings[key])

    if settings['Workflow folder'] == '':
        default_save_folder_dlg = wx.MessageDialog(None,
                                                   'Please select the default directory where Workflow files should be saved.',
                                                   'Choose default save location', wx.YES_NO | wx.ICON_INFORMATION)
        default_save_folder_dlg.SetYesNoLabels('Select', 'Exit')  # rename 'Yes' and 'No' labels to 'Select' and 'Exit'

        if default_save_folder_dlg.ShowModal() == wx.ID_YES:
            settings['Workflow folder'] = directory_chooser(None)
            save_settings(settings)
        else:
            raise SystemExit

        default_save_folder_dlg.Destroy()

    return settings


def import_settings():
    # open data/settings.json file to import settings, otherwise create empty dictionary to use factory settings
    try:
        with open('data/settings.json', 'r') as json_file:
            imported_settings = validate_settings(json.load(json_file))

    except (FileNotFoundError, json.decoder.JSONDecodeError) as error:
        if isinstance(error, FileNotFoundError):
            wx.MessageDialog(None, 'The \'settings.json\' file could not be located and has been reconstructed.',
                             'Missing settings.json file', wx.OK | wx.ICON_INFORMATION).ShowModal()
        if isinstance(error, json.decoder.JSONDecodeError):
            wx.MessageDialog(None, 'The \'settings.json\' file could not be decoded and has been reconstructed.',
                             'Corrupt settings.json file', wx.OK | wx.ICON_INFORMATION).ShowModal()

        # reconstruct settings.json file with factory settings
        with open('data/settings.json', 'w') as json_file:
            imported_settings = validate_settings(dict())
            json.dump(imported_settings, json_file, indent=4)

    return imported_settings


def save_settings(settings):
    settings = validate_settings(settings)
    with open('data/settings.json', 'w') as json_file:
        json.dump(settings, json_file, indent=4)


def open_settings(parent_window):
    settings_dlg = SettingsDialog(parent_window)
    if settings_dlg.ShowModal() == wx.ID_OK:
        settings_old = import_settings()
        save_settings(settings_dlg.settings)

        # prompt user to restart Aldras if parameters were changed affecting SelectionFrame or EditFrame
        difference = False
        for parameter in ['Number of recent workflows displayed', 'Workflow folder', 'Number of hotkeys',
                          'Large lines number',
                          'Notifications']:
            if settings_old[parameter] != settings_dlg.settings[parameter]:
                difference = True
                break

        if difference:
            settings_restart_dlg = wx.MessageDialog(settings_dlg,
                                                    'Aldras may need to be restarted for changes to be fully applied',
                                                    'Restart Aldras to apply changes', wx.YES_NO | wx.ICON_WARNING)
            settings_restart_dlg.SetYesNoLabels('Restart', 'Later')  # rename 'Yes' and 'No' labels to 'Restart' and 'Later'

            if settings_restart_dlg.ShowModal() == wx.ID_YES:
                # relaunch SelectionFrame
                if parent_window.GetName() == 'edit_frame':  # close EditFrame and save workflow if needed
                    parent_window.Close()
                    parent_window.parent.restart()
                elif parent_window.GetName() == 'selection_frame':
                    parent_window.restart()

            settings_restart_dlg.Destroy()

    settings_dlg.Destroy()


class SettingsDialog(wx.Dialog):
    """Main frame to select workflow."""

    def __init__(self, parent):
        wx.Dialog.__init__(self, parent, title=f'Settings', name='settings_frame')
        self.SetBackgroundColour('white')
        if parent:
            self.SetIcon(wx.Icon(parent.software_info.icon, wx.BITMAP_TYPE_ICO))  # assign icon

        margin = 30

        def setting_cb(parameter, parent_window=None):
            if not parent_window:
                parent_window = panel

            return wx.ComboBox(parent_window, value=str(self.settings[parameter]),
                               choices=settings_possibilities[parameter],
                               style=wx.CB_READONLY)

        self.settings = import_settings()
        self.settings_as_imported = self.settings.copy()

        static_boxsizer_inner_padding = 5
        static_boxsizer_outer_spacing = 12

        # create sizers
        vbox_outer = wx.BoxSizer(wx.VERTICAL)
        vbox_main = wx.BoxSizer(wx.VERTICAL)
        vbox_container = wx.BoxSizer(wx.VERTICAL)

        panel = wx.Panel(self)

        # add rescaled image
        png = wx.Image('data/settings.png', wx.BITMAP_TYPE_PNG).Scale(120, 120,
                                                                      quality=wx.IMAGE_QUALITY_HIGH)
        self.logo_img = wx.StaticBitmap(panel, wx.ID_ANY, wx.Bitmap(png))
        vbox_container.Add(self.logo_img, 0, wx.ALIGN_CENTER_HORIZONTAL | wx.NORTH | wx.SOUTH, round(margin / 2))

        workflow_folder_sizer = wx.StaticBoxSizer(wx.StaticBox(panel, wx.ID_ANY, 'Workflow Folder'), wx.VERTICAL)  # ---
        workflow_folder_hbox = wx.BoxSizer(wx.HORIZONTAL)

        self.workflow_folder_st = wx.StaticText(panel, label='')
        workflow_folder_hbox.Add(self.workflow_folder_st, 0, wx.ALIGN_CENTER_VERTICAL | wx.EAST, 10)
        self.set_workflow_folder_text()

        workflow_folder_hbox.AddStretchSpacer()

        workflow_folder_chooser = wx.DirPickerCtrl(panel, path=self.settings['Workflow folder'],
                                                   style=wx.DIRP_DIR_MUST_EXIST)
        workflow_folder_chooser.Bind(wx.EVT_DIRPICKER_CHANGED,
                                     lambda event: self.setting_change(event.GetPath(), 'Workflow folder'))
        workflow_folder_hbox.Add(workflow_folder_chooser, 0, wx.ALIGN_CENTER_VERTICAL)

        workflow_folder_sizer.Add(workflow_folder_hbox, 0, wx.EXPAND | wx.ALL, static_boxsizer_inner_padding)
        vbox_container.Add(workflow_folder_sizer, 0, wx.EXPAND | wx.SOUTH, static_boxsizer_outer_spacing)  # -----------

        #

        selection_sizer = wx.StaticBoxSizer(wx.StaticBox(panel, wx.ID_ANY, 'Workflow Selection'), wx.VERTICAL)  # ------

        num_recent_workflows_hbox = wx.BoxSizer(wx.HORIZONTAL)
        num_recent_workflows_st = wx.StaticText(panel, label='Number of recent workflows displayed:')
        num_recent_workflows_hbox.Add(num_recent_workflows_st, 0, wx.ALIGN_CENTER_VERTICAL | wx.EAST, 10)
        num_recent_workflows_cb = setting_cb('Number of recent workflows displayed')
        num_recent_workflows_cb.Bind(wx.EVT_COMBOBOX, lambda event: self.setting_change(event.GetString(),
                                                                                        'Number of recent workflows displayed'))
        num_recent_workflows_hbox.Add(num_recent_workflows_cb, 0, wx.ALIGN_CENTER_VERTICAL)

        selection_sizer.Add(num_recent_workflows_hbox, 0, wx.ALL, static_boxsizer_inner_padding)
        vbox_container.Add(selection_sizer, 0, wx.EXPAND | wx.SOUTH, static_boxsizer_outer_spacing)  # -----------------

        #

        mouse_monitor_sizer = wx.StaticBoxSizer(wx.StaticBox(panel, wx.ID_ANY, 'Mouse Monitor'), wx.VERTICAL)  # -------

        mouse_monitor_freeze_mthd_hbox = wx.BoxSizer(wx.HORIZONTAL)

        mouse_monitor_freeze_mthd_st = wx.StaticText(panel, label='Freeze method:')
        mouse_monitor_freeze_mthd_hbox.Add(mouse_monitor_freeze_mthd_st, 0, wx.ALIGN_CENTER_VERTICAL | wx.EAST, 10)

        mouse_monitor_freeze_mthd_cb = setting_cb('Freeze method')
        mouse_monitor_freeze_mthd_cb.Bind(wx.EVT_COMBOBOX,
                                          lambda event: self.setting_change(event.GetString(), 'Freeze method'))
        mouse_monitor_freeze_mthd_hbox.Add(mouse_monitor_freeze_mthd_cb, 0, wx.ALIGN_CENTER_VERTICAL)

        mouse_monitor_sizer.Add(mouse_monitor_freeze_mthd_hbox, 0, wx.ALL, static_boxsizer_inner_padding)
        vbox_container.Add(mouse_monitor_sizer, 0, wx.EXPAND | wx.SOUTH, static_boxsizer_outer_spacing)  # -------------

        #

        editor_collpane = wx.CollapsiblePane(panel,
                                             label=' Editor')  # -------------------------------------------------
        editor_collpane.GetChildren()[0].SetBackgroundColour(wx.WHITE)  # set button and label background
        editor_collpane.GetChildren()[0].Bind(wx.EVT_KILL_FOCUS,
                                              lambda event: None)  # prevents flickering when focus is killed
        vbox_container.Add(editor_collpane, 0, wx.GROW | wx.SOUTH, static_boxsizer_outer_spacing)
        editor_panel = editor_collpane.GetPane()

        editor_sizer = wx.StaticBoxSizer(wx.StaticBox(editor_panel), wx.VERTICAL)

        editor_number_of_hotkeys_hbox = wx.BoxSizer(wx.HORIZONTAL)

        editor_number_of_hotkeys_st = wx.StaticText(editor_panel, label='Number of hotkeys:')
        editor_number_of_hotkeys_hbox.Add(editor_number_of_hotkeys_st, 0, wx.ALIGN_CENTER_VERTICAL | wx.EAST, 10)
        editor_number_of_hotkeys_cb = setting_cb('Number of hotkeys', editor_panel)
        editor_number_of_hotkeys_cb.Bind(wx.EVT_COMBOBOX,
                                         lambda event: self.setting_change(event.GetString(), 'Number of hotkeys'))
        editor_number_of_hotkeys_hbox.Add(editor_number_of_hotkeys_cb, 0, wx.ALIGN_CENTER_VERTICAL)
        editor_sizer.Add(editor_number_of_hotkeys_hbox, 0, wx.ALL, static_boxsizer_inner_padding)

        editor_number_many_lines_hbox = wx.BoxSizer(wx.HORIZONTAL)
        editor_number_many_lines_st = wx.StaticText(editor_panel,
                                                    label='Number of large number of lines to prompt warning:')
        editor_number_many_lines_hbox.Add(editor_number_many_lines_st, 0, wx.ALIGN_CENTER_VERTICAL | wx.EAST, 10)
        editor_number_many_lines_cb = wx.TextCtrl(editor_panel, style=wx.TE_CENTRE,
                                                  value=str(self.settings['Large lines number']), size=wx.Size(40, -1),
                                                  validator=CharValidator('only_integer', self))
        editor_number_many_lines_cb.SetMaxLength(3)
        editor_number_many_lines_cb.Bind(wx.EVT_TEXT,
                                         lambda event: self.setting_change(event.GetString(), 'Large lines number'))
        editor_number_many_lines_hbox.Add(editor_number_many_lines_cb, 0, wx.ALIGN_CENTER_VERTICAL)
        editor_sizer.Add(editor_number_many_lines_hbox, 0, wx.ALL, static_boxsizer_inner_padding)

        editor_notifications_hbox = wx.BoxSizer(wx.HORIZONTAL)
        editor_notifications_st = wx.StaticText(editor_panel, label='Notifications:')
        editor_notifications_hbox.Add(editor_notifications_st, 0, wx.ALIGN_CENTER_VERTICAL | wx.EAST, 10)
        editor_notifications_cb = setting_cb('Notifications', editor_panel)
        editor_notifications_cb.Bind(wx.EVT_COMBOBOX,
                                     lambda event: self.setting_change(event.GetString(), 'Notifications'))
        editor_notifications_hbox.Add(editor_notifications_cb, 0, wx.ALIGN_CENTER_VERTICAL)
        editor_sizer.Add(editor_notifications_hbox, 0, wx.ALL, static_boxsizer_inner_padding)

        editor_panel.SetSizer(editor_sizer)
        editor_sizer.SetSizeHints(editor_panel)  # ---------------------------------------------------------------------

        #

        record_collpane = wx.CollapsiblePane(panel, label=' Record Options')  # ----------------------------------------
        record_collpane.GetChildren()[0].SetBackgroundColour(wx.WHITE)  # set button and label background
        record_collpane.GetChildren()[0].Bind(wx.EVT_KILL_FOCUS,
                                              lambda event: None)  # prevents flickering when focus is killed
        vbox_container.Add(record_collpane, 0, wx.GROW | wx.SOUTH, static_boxsizer_outer_spacing)
        record_panel = record_collpane.GetPane()

        record_sizer = wx.StaticBoxSizer(wx.StaticBox(record_panel), wx.VERTICAL)

        from modules.aldras_record import create_record_options
        record_options_sizer = create_record_options(record_panel, settings_frame=True)

        # bind parameter changes to setting_change()
        for record_pause_option in settings_possibilities['Record pause']:
            self.FindWindowByLabel(record_pause_option).Bind(wx.EVT_RADIOBUTTON, lambda event: self.setting_change(
                event.GetEventObject().GetLabel(), 'Record pause'))

        self.FindWindowByLabel(self.settings['Record pause']).SetValue(True)

        self.FindWindowByName('some_sleep_thresh').SetValue(str(self.settings['Record pause over duration']))
        self.FindWindowByName('some_sleep_thresh').Bind(wx.EVT_TEXT, lambda event: self.setting_change(
            event.GetEventObject().GetValue(), 'Record pause over duration'))

        self.FindWindowByName('Record method').SetSelection(
            self.FindWindowByName('Record method').FindString(self.settings['Record method']))
        self.FindWindowByName('Record method').Bind(wx.EVT_RADIOBOX, lambda event: self.setting_change(
            event.GetEventObject().GetString(event.GetEventObject().GetSelection()), 'Record method'))

        record_sizer.Add(record_options_sizer, 0, wx.ALL, static_boxsizer_inner_padding)
        record_panel.SetSizer(record_sizer)
        record_sizer.SetSizeHints(record_panel)  # ---------------------------------------------------------------------

        #

        execute_collpane = wx.CollapsiblePane(panel,
                                              label=' Execute Options')  # ----------------------------------------
        execute_collpane.GetChildren()[0].SetBackgroundColour(wx.WHITE)  # set button and label background
        execute_collpane.GetChildren()[0].Bind(wx.EVT_KILL_FOCUS,
                                               lambda event: None)  # prevents flickering when focus is killed
        vbox_container.Add(execute_collpane, 0, wx.GROW | wx.SOUTH, static_boxsizer_outer_spacing)
        execute_panel = execute_collpane.GetPane()

        execute_sizer = wx.StaticBoxSizer(wx.StaticBox(execute_panel), wx.VERTICAL)

        from modules.aldras_execute import create_execute_options
        execute_options_sizer = create_execute_options(execute_panel, settings_frame=True)

        for setting_name in ['Execute pause between commands', 'Execute pause between commands checked',
                             'Execute mouse command duration', 'Execute mouse command duration checked',
                             'Interval between text character outputs',
                             'Interval between text character outputs checked']:
            widget = self.FindWindowByName(setting_name)

            if isinstance(widget, wx.CheckBox):  # set true or false, not string
                widget.SetValue(self.settings[setting_name])
                widget.Bind(wx.EVT_CHECKBOX,
                            lambda event, setting=setting_name: self.setting_change(event.GetEventObject().GetValue(),
                                                                                    setting))
            elif isinstance(widget, wx.TextCtrl):
                widget.SetValue(str(self.settings[setting_name]))
                widget.Bind(wx.EVT_TEXT,
                            lambda event, setting=setting_name: self.setting_change(event.GetEventObject().GetValue(),
                                                                                    setting))

        execute_sizer.Add(execute_options_sizer, 0, wx.ALL, static_boxsizer_inner_padding)

        execute_panel.SetSizer(execute_sizer)
        execute_sizer.SetSizeHints(execute_panel)  # -------------------------------------------------------------------

        #

        panel.SetSizer(vbox_container)
        vbox_main.Add(panel, 0, wx.EXPAND)

        # add buttons
        btns = self.CreateSeparatedButtonSizer(wx.OK | wx.CANCEL)
        vbox_main.Add(btns, 0, wx.EXPAND | wx.ALL, 5)

        self.FindWindowById(wx.ID_OK).Enable(False)

        vbox_outer.AddStretchSpacer()
        vbox_outer.Add(vbox_main, 0, wx.EXPAND | wx.WEST | wx.EAST, margin)
        vbox_outer.AddSpacer(margin / 2)
        vbox_outer.AddStretchSpacer()

        self.SetSizerAndFit(vbox_outer)

        self.Center()

    def setting_change(self, value, setting):
        self.settings[setting] = value
        self.settings = validate_settings(self.settings)

        if setting == 'Workflow folder':
            self.set_workflow_folder_text()

        if self.settings != self.settings_as_imported:
            self.FindWindowById(wx.ID_OK).Enable()  # enable OK button if changes
        else:
            self.FindWindowById(wx.ID_OK).Enable(False)  # disable OK button if no changes

    def set_workflow_folder_text(self):
        """Set label of workflow folder static text"""
        workflow_folder_text = self.settings['Workflow folder']

        # cut out middle characters if path is too long
        workflow_folder_text_char_limit = 45
        if len(workflow_folder_text) > workflow_folder_text_char_limit:
            workflow_folder_text_start = workflow_folder_text[:round(workflow_folder_text_char_limit / 2) - 3]
            workflow_folder_text_end = workflow_folder_text[-round(workflow_folder_text_char_limit / 2) + 2:]
            workflow_folder_text = f'{workflow_folder_text_start} ... {workflow_folder_text_end}'

        self.workflow_folder_st.SetLabel(workflow_folder_text)


def profiler(output_limit=None, save_file=False):
    """Decorator to profile functions"""

    def profiler_dec(func):
        from functools import wraps

        @wraps(func)
        def wrapper(*args, **kwargs):
            import cProfile
            import pstats

            prof = cProfile.Profile()
            ret_val = prof.runcall(func, *args, **kwargs)

            stats = pstats.Stats(prof)
            stats.sort_stats(pstats.SortKey.TIME)
            stats.print_stats(output_limit)

            if save_file:
                from datetime import datetime
                prof.dump_stats(f'{func.__name__} - {datetime.now():%m-%d-%Y %H:%M:%S}.profile')  # save profile file

            return ret_val

        return wrapper

    return profiler_dec


@profiler(save_file=False)
def func1(arg1, arg2):
    """Example function"""
    return arg1, arg2


if __name__ == "__main__":
    f1 = func1(131232, 192)


import math
from operator import itemgetter
from tkinter import *

window = Tk()
window.wm_title("Collage Optimizer")  # name window

unit_input = StringVar(window)
unit_input.set("mm")

Label(window, text="Minimum number of photos").grid(row=0, column=0)
num_min_photos_input = Entry(window, width=6)
num_min_photos_input.grid(row=0, column=1)

Label(window, text="Maximum number of photos").grid(row=1, column=0)
num_max_photos_input = Entry(window, width=6)
num_max_photos_input.grid(row=1, column=1)

Label(window, text="Photo dimensions (H x W)").grid(row=0, column=2)
photo_height_input = Entry(window, width=6)
photo_height_input.grid(row=0, column=3)
Label(window, text="x").grid(row=0, column=4)
photo_width_input = Entry(window, width=6)
photo_width_input.grid(row=0, column=5)

Label(window, text="Collage aspect ratio (H : W)").grid(row=1, column=2)
desire_collage_aspect_ratio_height_input = Entry(window, width=6)
desire_collage_aspect_ratio_height_input.grid(row=1, column=3)
Label(window, text=":").grid(row=1, column=4)
desire_collage_aspect_ratio_width_input = Entry(window, width=6)
desire_collage_aspect_ratio_width_input.grid(row=1, column=5)

Label(window, text="Units").grid(row=2, column=2)
OptionMenu(window, unit_input, "in", "ft", "mm", "cm", "m").grid(row=2, column=3)


def calc_collage():
    global num_max_photos
    global photo_height
    global photo_width
    global collage_aspect_ratio_list
    global desired_collage_h_to_w_ratio
    global canvas_widget
    unit_text = str(unit_input.get())
    num_photos_width = math.floor(num_max_photos / num_rows)
    collage_height = num_rows * photo_height
    collage_width = num_photos_width * photo_width
    collage_aspect_ratio_list.append(collage_height / collage_width)

    output_text_line_1 = "Organizing the photos into %i rows of %i photos utilizes %i of the maximum %i photos.\n" % (
        num_rows, num_photos_width, num_max_photos - (num_max_photos % num_rows), num_max_photos)
    output_text_line_2 = "The resulting collage is %.2f %s high %.2f %s wide.\n" % (collage_height, unit_text, collage_width, unit_text)
    output_text_line_3 = "Actual aspect ratio of %.2f compared to the desired %.2f." % (
        collage_height / collage_width, desired_collage_h_to_w_ratio)

    output_text_all_lines = output_text_line_1 + output_text_line_2 + output_text_line_3

    output_text = Text(window, width=85, height=3)
    output_text.insert(END, output_text_all_lines)
    output_text.grid(row=3, column=0, columnspan=100)

    Button(window, text="Add row", command=add_row).grid(row=4, column=0)
    Button(window, text="Subtract row", command=subtract_row).grid(row=4, column=1)

    vis_scale = 10
    vis_padding = 30
    canvas_width = int(3 * vis_padding + vis_scale * collage_width)
    canvas_height = int(vis_padding + vis_scale * collage_height)
    canvas_widget = Canvas(window, width=canvas_width, height=canvas_height)
    canvas_widget.grid(row=5, column=0, columnspan=100)

    for row in range(num_rows):
        for col in range(num_photos_width):
            start_x = 7 + (photo_width * col)
            start_y = 1 + (photo_height * row)
            end_x = start_x + photo_width
            end_y = start_y + photo_height

            canvas_widget.create_rectangle(vis_scale * start_x, vis_scale * start_y, vis_scale * end_x,
                                           vis_scale * end_y, fill="")

    canvas_widget.create_text(35, ((canvas_height - 40) / 2), text=str("%.2f %s" % (collage_height, unit_text)))
    canvas_widget.create_text((canvas_width + 45) / 2, canvas_height - 8, text=str("%.2f %s" % (collage_width, unit_text)))


def calc_button():
    global num_rows
    global num_max_photos
    global photo_height
    global photo_width
    global collage_aspect_ratio_list
    global desired_collage_h_to_w_ratio
    global canvas_widget

    try:
        num_min_photos = float(num_min_photos_input.get())
    except ValueError:
        popup_window_input = Tk()
        Label(popup_window_input, text="Please enter a valid minimum number of photos.").grid(row=0, column=0)
        popup_window_input.mainloop()

    try:
        num_max_photos = float(num_max_photos_input.get())
    except ValueError:
        popup_window_input = Tk()
        Label(popup_window_input, text="Please enter a valid maximum number of photos.").grid(row=0, column=0)
        popup_window_input.mainloop()

    try:
        photo_height = float(photo_height_input.get())
    except ValueError:
        popup_window_input = Tk()
        Label(popup_window_input, text="Please enter a valid number for the height of the photos.").grid(row=0, column=0)
        popup_window_input.mainloop()

    try:
        photo_width = float(photo_width_input.get())
    except ValueError:
        popup_window_input = Tk()
        Label(popup_window_input, text="Please enter a valid number for the width of the photos.").grid(row=0, column=0)
        popup_window_input.mainloop()

    try:
        desire_collage_aspect_ratio_height = float(desire_collage_aspect_ratio_height_input.get())
    except ValueError:
        popup_window_input = Tk()
        Label(popup_window_input, text="Please enter a valid number for the aspect height of the collage.").grid(row=0, column=0)
        popup_window_input.mainloop()

    try:
        desire_collage_aspect_ratio_width = float(desire_collage_aspect_ratio_width_input.get())
    except ValueError:
        popup_window_input = Tk()
        Label(popup_window_input, text="Please enter a valid number for the aspect height of the collage.").grid(row=0, column=0)
        popup_window_input.mainloop()

    desired_collage_h_to_w_ratio = desire_collage_aspect_ratio_height / desire_collage_aspect_ratio_width

    collage_aspect_ratio_list = []

    for num_rows in range(2, math.floor(num_max_photos / 2)):
        num_photos_width = math.floor(num_max_photos / num_rows)

        collage_height = num_rows * photo_height
        collage_width = num_photos_width * photo_width

        collage_aspect_ratio_list.append(collage_height / collage_width)

    collage_aspect_ratio_delta_list = []
    for collage_aspect_ratio in collage_aspect_ratio_list:
        collage_aspect_ratio_delta_list.append(abs(collage_aspect_ratio - desired_collage_h_to_w_ratio))

    num_rows = 2 + min(enumerate(collage_aspect_ratio_delta_list), key=itemgetter(1))[0]

    try:
        canvas_widget.delete("all")
    except NameError:
        pass


    calc_collage()


def add_row():
    global num_rows
    if num_rows != num_max_photos - 1:
        num_rows += 1
        canvas_widget.delete("all")
        calc_collage()


def subtract_row():
    global num_rows
    if num_rows != 1:
        num_rows -= 1
        canvas_widget.delete("all")
        calc_collage()


Button(window, text="Calculate", command=calc_button).grid(row=2, column=0)

mainloop()


import itertools
import os
import platform
import re
import time
from concurrent.futures import ThreadPoolExecutor

import pandas as pd
from docx import Document

if platform.system() == "Windows":
    import pythoncom
    from docx2pdf import convert


def docx_replace_regex(doc_obj: Document, regex_to_replace: re.compile, replacement: str) -> None:
    """
    Replace the regex in a docx.Document object.

    Args:
        doc_obj: Template `docx.Document` object.
        regex_to_replace: Regex compile to replace in the document.
        replacement : Replacement string to substitute for the regex.
    """
    for p in doc_obj.paragraphs:
        if regex_to_replace.search(p.text):
            inline = p.runs
            # Loop added to work with runs (strings with same style)
            for i in range(len(inline)):
                if regex_to_replace.search(inline[i].text):
                    text = regex_to_replace.sub(replacement, inline[i].text)
                    inline[i].text = text

    for table in doc_obj.tables:
        for row in table.rows:
            for cell in row.cells:
                docx_replace_regex(cell, regex_to_replace, replacement)


def filename_ext(filename: str) -> str:
    return os.path.splitext(filename)[1]


def generate_doc(
    doc_spec: tuple[int, pd.Series],
    template_docx: str,
    output_dir: str,
    output_base_fn: str,
    output_filetype: str,
) -> str:
    doc_number = doc_spec[0]
    doc_replacements = doc_spec[1]

    if platform.system() == "Windows":
        pythoncom.CoInitialize()  # needed to prevent docx2pdf multithreading error

    _log = []

    _log.append(f"Document {doc_number+1} - creating replacements for:")
    doc = Document(template_docx)
    output_fn_additions = []

    # Loop through each replacement in the row for the document
    for str_to_replace, replacement_str in doc_replacements.items():
        _log.append(f"\t{str_to_replace} -> {replacement_str}")

        # Add column values that are not excluded to list to be added to
        # output filename
        column_substrings_excl_from_fn = [
            "date",
            "industry",
        ]
        if not any(substr in str_to_replace.lower() for substr in column_substrings_excl_from_fn):
            output_fn_additions.append(replacement_str)

        # Execute document replacement
        regex = re.compile(re.escape(str_to_replace))
        docx_replace_regex(doc, regex, replacement_str)

    # Save .docx file
    output_fn_addition_str = " - ".join(output_fn_additions)
    if output_fn_addition_str:
        output_fn_addition_str = f" - {output_fn_addition_str}"
    output_fn = os.path.join(output_dir, f"{output_base_fn}{output_fn_addition_str}.docx")
    doc.save(output_fn)

    if output_filetype == ".pdf":
        # Convert .docx file to .pdf file
        convert(output_fn)
        os.remove(output_fn)  # remove .docx version
        output_fn = output_fn.replace(".docx", ".pdf")

    _log.append(f"Document {doc_number+1} - file saved to '{output_fn}'.")

    return output_fn


def batch_replace(
    *,
    template_docx: str,
    replacements_csv: str,
    max_new_docs: int = 25,
    output_dir: str,
    output_base_fn: str,
    output_filetype: str,
) -> list[str]:
    """
    Generate multiple output files by executing multiple sets of find-replace operations.

    Args:
        template_docx: Path to the template document.
        replacements_csv: Path to the replacements spec document. Each row
            specifies a document version to generate. Each column header
            specifies the text to replace in the template. The column values
            specify the text to substitute during the replacement.
        output_dir: Path to the output directory.
        output_base_fn: Base filename for the generated output files.
        output_filetype: Output filetype (.docx, .pdf)
    """

    replacements_df = pd.read_csv(replacements_csv)
    replacements_df = replacements_df.truncate(after=max_new_docs - 1)  # limit number of documents generated
    output_filetype = output_filetype if output_filetype in [".docx", ".pdf"] else ".pdf"

    # Check input file extensions
    if filename_ext(template_docx) != ".docx":
        raise ValueError(f"{template_docx} does not have a valid file extension.")
    if filename_ext(replacements_csv) != ".csv":
        raise ValueError(f"{replacements_csv} does not have a valid file extension.")

    start = time.time()

    # # Loop through each replacement row
    # # doc_spec is containerized in tuple to match ThreadPoolExecutor behavior
    # output_file_paths = []
    # for doc_number, doc_replacements in replacements_df.iterrows():
    #     output_file_paths.append(
    #         generate_doc(
    #             (doc_number, doc_replacements),
    #             template_docx,
    #             output_dir,
    #             output_base_fn,
    #             output_filetype,
    #         )
    #     )

    with ThreadPoolExecutor() as executor:
        output_file_paths = executor.map(
            generate_doc,
            replacements_df.iterrows(),
            itertools.repeat(template_docx),
            itertools.repeat(output_dir),
            itertools.repeat(output_base_fn),
            itertools.repeat(output_filetype),
        )

    print("Time elapsed:", time.time() - start)

    return output_file_paths


if __name__ == "__main__":
    output_file_paths = batch_replace(
        template_docx="cover_letter_test.docx",
        replacements_csv="replacements.csv",
        max_new_docs=5,
        output_dir="created",
        output_base_fn="Cover Letter Test",
        output_filetype=".docx",
    )

    from os.path import basename
    from zipfile import ZipFile

    with ZipFile("generated_documents.zip", "w") as zip_obj:
        for generated_doc_path in output_file_paths:
            zip_obj.write(generated_doc_path, basename(generated_doc_path))